@@ -90,7 +90,8 @@ class CoreServer(ServerProtocol):
         self.server._running_port = port_used
         if self.server._running_stage < 2:
             self.server._running_stage = 2
-            self.server.state.server_ready()
+            self.server.state.ready()
+            self.server.context.ready()
             if self.server.controller.on_server_ready.exists():
                 self.server.controller.on_server_ready(**self.server.state.to_dict())
 
@@ -150,10 +151,30 @@ class CoreServer(ServerProtocol):
         logger.action_s2c(actions)
         self.publish("trame.actions.topic", actions)
 
+    # ---------------------------------------------------------------
+    # Internal RPCs
+    # ---------------------------------------------------------------
+
+    def clear_state_client_cache(self, *keys):
+        for k in keys:
+            del self._clients_state[k]
+
     # ---------------------------------------------------------------
     # RPCs
     # ---------------------------------------------------------------
 
+    @exportRpc("trame.force.push")
+    def force_push_state(self, *keys):
+        state_to_send = {key: self.server.state[key] for key in keys}
+        if state_to_send:
+            self.publish(
+                "trame.state.topic",
+                state_to_send,
+                skip_last_active_client=False,
+            )
+
+    # ---------------------------------------------------------------
+
     @exportRpc("trame.lifecycle.update")
     def life_cycle_update(self, name):
         _fn = self.server.controller[f"on_{name}"]
@@ -182,8 +203,9 @@ class CoreServer(ServerProtocol):
     async def trigger(self, name, args, kwargs):
         logger.action_c2s({"name": name, "args": args, "kwargs": kwargs})
         with self.server.state:
-            if name in self.server._triggers:
-                result = self.server._triggers[name](*args, **kwargs)
+            fn = self.server.controller.trigger_fn(name)
+            if fn:
+                result = fn(*args, **kwargs)
                 if inspect.isawaitable(result):
                     result = await result
                 return result
